/*
 Standalone Voltage Logger for ATtiny85 (Mod for SATL)
 Version: 1.1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 5/11/2011
 Last Modified: 8/11/2011
 
 Used to monitor voltage every x minutes/hours which is saved 
 to the built in EEPROM. You can transfer the data by connecting via 
 USB to your computer. Maximum of 255 readings can be logged. 
 Measures 0 - 15V reliably. Voltage range can be increased if the 2 diodes are removed.
 
 Remember to change your fuse bits so the ATtiny85 uses 16MHz PLL and divide clock by 8
 avrdude -p ATtiny85 -c usbtiny -U lfuse:w:0x61:m -U hfuse:w:0xdf:m -U efuse:w:0xff:m 
 
 2.7V Brown-out Detection is optional and is only recommended for CR2032 batteries that 
 are new or near new condition (-U hfuse:w:0xdd:m)
 
 */

//                  +-\/-+
// Ain0 (D 5) PB5  1|o   |8  Vcc
// Ain3 (D 3) PB3  2|    |7  PB2 (D 2) Ain1
// Ain2 (D 4) PB4  3|    |6  PB1 (D 1) pwm1
//            GND  4|    |5  PB0 (D 0) pwm0
//                  +----+

int VoltagePin = 3; // Analog input 3
#define ledPin PB0 // Output
#define buttonPin PB4 // Input
#define usbminusPin PB1 // USB D-

float vRefperbit = 0.001074;
int vRatio = 27;

#define F_CPU 16500000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include <stdlib.h>
#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include "setup.c"

volatile boolean f_pin = 0; // Pin change variable
volatile boolean buttonPressed = 0; // Keep track of button presses and releases

int functionSelect = WAITINPUT;
int dataCount = 1;
int delayTime = -1;

// Send the next digit from the EEPROM address
int sendnextDigit(int eepromAddress) {
	uchar digit;
	
	// Re-construct 10bit voltValue (0 - 1024) from high and low byte from EEPROM
	byte highByte = eeprom_read_byte((uint8_t*) eepromAddress);
	byte lowByte = eeprom_read_byte((uint8_t*) (eepromAddress+1));
	
	unsigned int combinedValue = (highByte << 8) | lowByte; // Combine high and low byte
	
	if (combinedValue == 65278) { // 254 high and 254 low byte
		return 1;
	}
	if (combinedValue == 65535) { // 255 high and 255 low byte (testing USB communication)
		combinedValue = 0;
	}
	
	double value = (combinedValue * vRefperbit) * vRatio; // Calculate ADC value to voltage level
	
	char tempValue[7];
	dtostrf(value, 3, 2, tempValue); // 3 character (including dot) minimum, 2 digit precision
	
	nextDigit = &valueBuffer[sizeof(valueBuffer)];
	*--nextDigit = 0xff;
	*--nextDigit = 0;
	*--nextDigit = KEY_RETURN;
	
	// Break up the number and convert to USB HID values
	int x = 4;
	if (tempValue[1] == 46) { // 46 = ASCII Dot detected after 1 digit, only send 4 characters instead of 5
		x = 3;
	}
	while (x >= 0) {
		digit = tempValue[x] - 48;
		if (digit == 0){
			*--nextDigit = KEY_0;
		}
		else if (digit <= 9) {
			*--nextDigit = KEY_1 - 1 + digit;
		}
		else {
			*--nextDigit = KEY_DOT;
		}
		x--;
	}
	
	return 0;
}

// Transfers the EEPROM data to USB. Stops if EEPROM value is 0 or the end of the EEPROM is reached (512).
void transferData(void) {
	boolean inusbTransfer = true;
	int readCount = 1;
	while (inusbTransfer == true) {
		usbPoll();
		if(usbInterruptIsReady() && nextDigit != NULL){ // We can send another digit
			buildReport();
			usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
			if(*++nextDigit == 0xff) { // This was terminator character
				nextDigit = NULL;
			}
		}
		if (nextDigit == NULL) {
			if (readCount < 510) {
				int checkforEnd = sendnextDigit(readCount);
				if (checkforEnd == true) { // End if result returned is 1
					inusbTransfer = false;
				}
				readCount+= 2;
			}
			else {
				inusbTransfer = false;
			}
		}
	}
}

// Blink the LED x number of times with the pause calculated by the watchdog timer selected and skip the LED turn off delay if needed
void blinkLed (byte times, byte wdttime, byte skipLedoffdelay) {
	for (int x = 0; x < times; x++) {
		PORTB |= (1<<ledPin); // LED on
		setup_watchdog(wdttime);
		system_sleep();
		
		PORTB &= ~(1<<ledPin); // LED off	
		if (skipLedoffdelay == false) {
			setup_watchdog(wdttime);
			system_sleep();
		}
	}
	turnOffwatchdog();
}

int main(void) {
	setup();
	
	DDRB |= (1<<ledPin); // Set Outputs
	DDRB &= ~((1<<buttonPin) | (1<<usbminusPin)); // Set Inputs
	
	// Pin interrupt setup
	sbi(GIMSK,PCIE); // Enable pin change interrupt
	sbi(PCMSK,PCINT4); // Apply interrupt to button pin
	
	while(1) {
		// Wait for input
		if (functionSelect == WAITINPUT) {
			
			// Read delay time
			if (delayTime == -1) {
				delayTime = eeprom_read_byte((uint8_t*) 0);
				if (delayTime > 10) { // Not a normal value
					delayTime = 0;
				}
			}
			
			// Turn off LED and reset buttonPressed before going to sleep
			PORTB &= ~(1<<ledPin);
			buttonPressed = 0;
			system_sleep();
			
			// Button was pressed, first test we are connected via USB by checking if the D- pin is high
			int usbConnected = PINB & (1<<usbminusPin);
			if (usbConnected) { // Send data from EEPROM to USB
				PORTB |= (1<<ledPin);
				
				// Increase clock speed to 16MHz
				CLKPR = (1<<CLKPCE); // Prescaler enable
				CLKPR = 0; // Clock division factor 1 (0000)
				
				startUSB();
				transferData();
				
				// Decrease clock speed to 1MHz
				CLKPR = (1<<CLKPCE); // Prescaler enable
				CLKPR = (1<<CLKPS2); // Clock division factor 16 (0100)
				
				PORTB &= ~(1<<ledPin);
			}
			else { // Not connected via USB, continue to normal functions
				
				setup_watchdog(T2S); // Sleep for 2 seconds
				system_sleep();
				if (buttonPressed != 1) { // Was the button released?
					int newbutton = buttonPressed;
					while (true) { // Sleep for 1 second after each button press, if no button press was made then exit
						setup_watchdog(T1S);
						system_sleep();
						if (newbutton != buttonPressed) {
							newbutton = buttonPressed;
						}
						else {
							break;
						}
					}
				}
				turnOffwatchdog();
				
				// User held down the button
				if (buttonPressed == 1) { // Change delay time
					cbi(GIMSK,PCIE); // Disable pin change interrupt (so that interrupt is not triggered when user releases button)
					functionSelect = CONFIGDELAY;
					delayTime = -1;
					blinkLed(1, T2S, SKIPLEDOFFDELAY);
					sbi(GIMSK,PCIE); // Re-enable pin change interrupt
				}
				
				// Start logging
				else if (buttonPressed == 6) { // Check if 6, as buttonPressed is incremented when user pushes and releases button
					functionSelect = STARTLOGGING;
					dataCount = 1;
					blinkLed(3, T500MS, 0);
				}
			}
		}
		
		// Log Voltage
		if (functionSelect == STARTLOGGING) {
			int voltValue = (analogRead(VoltagePin));
			
			// Split 10bit voltValue (0 - 1024) to high and low byte and write to EEPROM
			if (dataCount < 510) {
				eeprom_write_byte((uint8_t*) dataCount, (voltValue >> 8));
				dataCount++;
				eeprom_write_byte((uint8_t*) dataCount, (voltValue & 0xFF));
				dataCount++;
				if (dataCount < 510) {
					eeprom_write_byte((uint8_t*) dataCount, 254); // So we know where to stop when extracting data
					eeprom_write_byte((uint8_t*) (dataCount+1), 254);
				}
			}
			else {
				functionSelect = WAITINPUT;
			}
			
			// Calculate the delay time
			int waitCounter = 0;
			int waitTime = 0;
			f_pin = 0;
			if (delayTime == 0) {
				waitTime = 15;
			}
			if (delayTime >= 1 && delayTime < 4) {
				waitTime = (delayTime * 5) * 15;
			}
			if (delayTime >= 4 && delayTime < 6) {
				waitTime = ((delayTime-3) * 30) * 15;
			}
			if (delayTime >= 6 && delayTime < 8) {
				waitTime = ((delayTime-5) * 120) * 15;
			}
			if (delayTime == 8) {
				waitTime = 8 * 60 * 15;
			}
			if (delayTime == 9) {
				waitTime = 12 * 60 * 15;
			}
			if (delayTime == 10) {
				waitTime = 24 * 60 * 15;
			}
			
			while (waitCounter != waitTime) {
				setup_watchdog(T4S);
				system_sleep();
				waitCounter++;
				if (f_pin == 1) { // Did a pin change wake us up?
					cbi(GIMSK,PCIE); // Disable pin change interrupt
					functionSelect = WAITINPUT;
					turnOffwatchdog(); // Disable the watchdog
					waitCounter = waitTime; // Break out of the loop
					blinkLed(2, T500MS, 0);
					sbi(GIMSK,PCIE); // Re-enable pin change interrupt
				}
			}
		}
		
		// Configure the delay time
		if (functionSelect == CONFIGDELAY) {
			system_sleep(); // Wait for next button press
			f_pin = 0;
			
			setup_watchdog(T1S);
			system_sleep();
			turnOffwatchdog();
			
			if (f_pin == 1) { // Button let go
				delayTime++;
				blinkLed(1, T500MS, SKIPLEDOFFDELAY);
				if (delayTime > 10) { // Incorrect value
					delayTime = -1;
					blinkLed(1, T2S, SKIPLEDOFFDELAY);
				}
			}
			
			// User held down the button to save setting
			else if (f_pin == 0) { 
				cbi(GIMSK,PCIE); // Disable pin change interrupt (so that interrupt is not triggered when user releases button)
				if (delayTime >= 0 && delayTime < 10) { // Save to EEPROM
					functionSelect = WAITINPUT;
					eeprom_write_byte((uint8_t*)0, delayTime);
					blinkLed(3, T500MS, 0);
				}
				else { // Incorrect value
					delayTime = -1;
					blinkLed(1, T2S, SKIPLEDOFFDELAY);
				}
				sbi(GIMSK,PCIE); // Re-enable pin change interrupt
			}
		}
	}
	return 0;
}

ISR(WDT_vect) { }

ISR(PCINT0_vect) {
	f_pin = 1; // Set global button pressed (or released) flag
	buttonPressed++;
}
